##
# Copyright (c) 2008 Guido Guenther <agx@sigxcpu.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##

"""
NSS Directory service interfaces.

Uses libc's Name Service Switch for user and groups (/etc/nsswitch.conf).
"""

__all__ = [
    "NssDirectoryService",
    "NssDirectoryServiceCalDAV",
    "NssDirectoryServiceCardDAV",
    "CheckAndSetGUID"
]

from twistedcaldav.directory.cachingdirectory import CachingDirectoryService,\
    CachingDirectoryRecord
from twisted.cred.credentials import UsernamePassword
from twistedcaldav.directory.util import uuidFromName
from txdav.caldav.datastore.scheduling.cuaddress import normalizeCUAddr
from twisted.python import log
from twistedcaldav.directory.SynoShortNameGUIDMapping import ShortNameGUIDMapping
import pwd, grp, socket
import PAM
import time

GUID_CardDAVServer = "F3B04844-A3F3-5419-81D6-A5E14AC8F440" #F3B... is guid generate for nss with realm name "CardDAVServer"
Mapping_DB_Path = "/var/packages/CardDAVServer/target/var/db/ShortNameGUIDMapping.db"

class NsSwitch(object):
    """Simple interface to the nsswitch calls"""

    def get_user(self, username):
        try:
            return pwd.getpwnam(username)
        except KeyError:
            return None

    def get_group(self, groupname):
        try:
            return grp.getgrnam(groupname)
        except KeyError:
            return None

    def get_users(self):
        return pwd.getpwall()

    def get_groups(self):
        return grp.getgrall()


class NssDirectoryService(CachingDirectoryService):
    """
    Nss based Directory Service of L{IDirectoryService}
    """

    proj = ""
    baseGUID = "8EFFFAF1-5221-4813-B971-58506B963573"

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.realmName)

    def __init__(self, params):
        """
        @param params: a dictionary containing the following keys:
            cacheTimeout, realmName, groupPrefix, mailDomain, firstValidUid,
            lastValidUid, firstValidGid, lastValidGid
        """

        defaults = {
            "realmName": "Test Realm",
            # we only consider groups starting with:
            "groupPrefix": "caldavd-",
            # dont set calendarUserAdresses by default
            "mailDomain": None,
            # exclude system users and nobody by "default":
            "firstValidUid": 1000,
            "lastValidUid": 65533,
            "firstValidGid": 1000,
            "lastValidGid": 65533,
            "cacheTimeout": 1, # Minutes
            "augmentService" : None,
            "groupMembershipCache" : None,
        }
        ignored = None
        params = self.getParams(params, defaults, ignored)

        super(NssDirectoryService, self).__init__(params['cacheTimeout'])

        self.nsswitch = NsSwitch()
        self.realmName = params["realmName"]
        self.mailDomain = params["mailDomain"]
        self.groupPrefix = params["groupPrefix"]
        self.first_valid_uid = params["firstValidUid"]
        self.first_valid_gid = params["firstValidGid"]
        self.last_valid_uid = params["lastValidUid"]
        self.last_valid_gid = params["lastValidGid"]
        self.augmentService = params["augmentService"]
        self.groupMembershipCache = params["groupMembershipCache"]

    def recordTypes(self):
        recordTypes = (
            self.recordType_users,
            self.recordType_groups,
        )
        return recordTypes

    def _isValidUid(self, uid):
        if uid >= self.first_valid_uid and uid <= self.last_valid_uid:
            return True

    def _isValidGid(self, gid):
        if gid >= self.first_valid_gid and gid <= self.last_valid_gid:
            return True

    def queryDirectory(self, recordTypes, indexType, indexKey):
        self.log.debug("Querying directory for recordTypes %s, "
                       "indexType %s and indexKey %s" %
                       (recordTypes, indexType, indexKey),
                       system="NssDirectoryService")

        def _recordWithGUID(recordType, guid):
            # Code has to be written to query on GUID
            self.log.debug("syno - in new GUID func")
            record = NssGUIDRecord(
                            service = self,
                            guid = guid,
                            proj = self.proj
                            )
            return record

        def _recordWithShortName(recordType, shortName):
            record = None
            if recordType == self.recordType_users:
                self.log.debug("syno - user type in shortname: %s" % shortName)
                #establish shoartnam <-> GUID mapping at first time
                try:
                    buf = shortName.decode('utf-8') #try decode unicode object
                except UnicodeDecodeError: #unicode string content in pure str obj, ex: "m\xf6ts"
                    buf="";
                    for i in shortName:
                        buf+=unichr(ord(i)).encode('utf-8')#traverse all char and turn it to unicode and convert it back to str obj
                    shortName = buf

                self.log.debug("syno - username: %s shortName" % shortName)
                result = self.nsswitch.get_user(shortName)
                if result and self._isValidUid(result[2]):
                    self.log.debug("syno - Found it: %s %s" % (result[0], result[4]))
                    record = NssUserRecord(
                                service = self,
                                userName = result[0],
                                gecos = result[4],
                                proj = self.proj
                             )
            elif recordType == self.recordType_groups:
                result = self.nsswitch.get_group(self.groupPrefix + shortName)
                if result: # and self._isValidGid(result[2]): FIXME
                    record = NssGroupRecord(
                                service = self,
                                groupName = result[0],
                                members = result[3]
                             )
            return record

        for recordType in recordTypes:
            record = None
            if indexType == self.INDEX_TYPE_GUID:
                record = _recordWithGUID(recordType, indexKey)
            elif indexType == self.INDEX_TYPE_SHORTNAME:
                record = _recordWithShortName(recordType, indexKey)
            elif indexType == self.INDEX_TYPE_CUA:
                address = normalizeCUAddr(indexKey)
                if address.startswith("urn:uuid:"):
                    guid = address[9:]
                    record = _recordWithGUID(recordType, guid)
                elif address.startswith("mailto:") and \
                     address.endswith("@"+self.mailDomain):
                    shortName = address[7:].partition("@")[0]
                    record = _recordWithShortName(recordType, shortName)
            elif indexType == self.INDEX_TYPE_AUTHID:
                pass

            if record:
                if self.augmentService is not None:
                    d = self.augmentService.getAugmentRecord(record.guid,
                                                             record.recordType)
                    d.addCallback(lambda x:record.addAugmentInformation(x))

                self.recordCacheForType(recordType).addRecord(
                        record, indexType, indexKey
                )

                # We got a match, so don't bother checking other types
                break

class NssDirectoryServiceCalDAV(NssDirectoryService):
    proj = "CalDAV"

class NssDirectoryServiceCardDAV(NssDirectoryService):
    proj = "CardDAV"

class NssDirectoryRecord(CachingDirectoryRecord):
    """
    Nss Directory Record
    """
    def __init__(self, service, recordType, shortNames, guid=None,
                 fullName=None, emailAddresses=set()):
        super(NssDirectoryRecord, self).__init__(
            service               = service,
            recordType            = recordType,
            guid                  = guid,
            shortNames            = shortNames,
            fullName              = fullName,
            emailAddresses        = emailAddresses,
        )


class NssGUIDRecord(NssDirectoryRecord):
    def __init__(self, service, guid, proj):
        recordType = service.recordType_users
        mapping = ShortNameGUIDMapping(proj)
        shortNames = (mapping.get_shortname_from_GUID(guid), )
        super(NssGUIDRecord, self).__init__(service, recordType, shortNames,guid=guid)

class NssUserRecord(NssDirectoryRecord):
    """
    NSS Users implementation of L{IDirectoryRecord}.
    """
    def __init__(self, service, userName, gecos, proj):
        recordType = service.recordType_users
        shortNames = (userName,)
        fullName = gecos.split(",",1)[0]
        emailAddresses = set()
        self.proj = proj
        guid = uuidFromName(GUID_CardDAVServer, "users:%s" % shortNames) #F3B... is guid generate for nss with realm name "CardDAVServer"
        mapping = ShortNameGUIDMapping(proj)
        if mapping.is_GUID_exist(guid) != 1:
            userName=userName.decode('utf-8');#turn username to unicode for sqlite
            mapping.insert_shortname_GUID_mapping(userName, guid)
        if service.mailDomain:
            emailAddresses.add("%s@%s" % (userName, service.mailDomain))
        super(NssUserRecord, self).__init__(service, recordType, shortNames, guid=guid,
                                            fullName=fullName,
                                            emailAddresses=emailAddresses)

    def groups(self):
        for result in self.service.nsswitch.get_groups():
            if self.service._isValidGid(result[2]) and \
               result[0].startswith(self.service.groupPrefix) and \
               self.shortNames[0] in result[3]:
                yield self.service.recordWithShortName(
                        self.service.recordType_groups,
                        result[0].replace(self.service.groupPrefix,'',1)
                      )

    def getIP(self, logPath, timeInfo):
        ip = ""
        with open(logPath) as inFile:
            fdata = inFile.readlines()

        for line in fdata[-50:]:
            msgParsed = line.split()

            #2016-02-03 14:48:54+0800 [-] [caldav-1]  [HTTPChannel,20,::ffff:192.168.60.194] [twistedcaldav.directory.nss.NssUserRecord#debug] 1454482134.76
            if len(msgParsed) == 7 and msgParsed[-1] == timeInfo :
                ## [HTTPChannel,20,::ffff:192.168.60.194]
                ip = msgParsed[4].split(",")[-1][:-1]
                if ip.split(":")[0] == "":
                    ip = ip.split(":")[-1]
                break

        return ip

    def verifyCredentials(self, credentials):
        if isinstance(credentials, UsernamePassword):
            # Check that the username supplied matches the shortName
            # (The DCS might already enforce this constraint, not sure)
            try:
                buf = credentials.username.decode('utf-8')
            except UnicodeDecodeError:
                buf="";
                for i in credentials.username:
                    buf+=unichr(ord(i)).encode('utf-8')
                credentials.username = buf
            timeInfo = str(time.time())
            self.log.debug("%s" % (timeInfo))
            self.log.debug("syno - PAM username %s"% credentials.username)
            if credentials.username not in self.shortNames:
                return False

            # Check cached password
            try:
                if "" != self.password and credentials.password == self.password:
                    return True
            except AttributeError:
                pass

            # Authenticate against PAM
            def pam_conv(auth, query_list, userData):
                return [(credentials.password, 0)]

            auth = PAM.pam()
            if self.proj == "CalDAV":
                auth.start("caldav")
                logPath="/var/packages/CalDAVServer/target/var/log/caldavd/error.log"
            else:
                auth.start("carddav")
                logPath="/var/packages/CardDAVServer/target/var/log/caldavd/error.log"

            connectionIP = self.getIP(logPath, timeInfo)
            auth.set_item(PAM.PAM_RHOST, connectionIP)

            auth.set_item(PAM.PAM_USER, credentials.username)
            auth.set_item(PAM.PAM_CONV, pam_conv)
            try:
                auth.authenticate()
            except PAM.error, resp:
                self.password = "";
                return False
            else:
                # Cache the password to avoid future DS queries
                self.password = credentials.password
                return True

        return super(NssUserRecord, self).verifyCredentials(credentials)


class NssGroupRecord(NssDirectoryRecord):
    """
    NSS Groups implementation of L{IDirectoryRecord}.
    """
    def __init__(self, service, groupName, members=()):
        recordType = service.recordType_groups
        shortNames  = (groupName.replace(service.groupPrefix,'',1),)
        super(NssGroupRecord, self).__init__(service, recordType, shortNames)
        self._members = members

    def members(self):
        for shortName in self._members:
            yield self.service.recordWithShortName(
                    self.service.recordType_users,
                    shortName
                  )

class CheckAndSetGUID:
    def __init__(self, proj, userName):
        guid = uuidFromName(GUID_CardDAVServer, "users:%s" % userName) #F3B... is guid generate for nss with realm name "CardDAVServer"
        mapping = ShortNameGUIDMapping(proj)
        if mapping.is_GUID_exist(guid) != 1:
            userName=userName.decode('utf-8');#turn username to unicode for sqlite
            mapping.insert_shortname_GUID_mapping(userName, guid)
        print guid

